(* dataspec.sml
 * 
 * Main file for dataspec
 *)


structure DataSpec= struct
  
  structure A = Atom
  
  structure AtomPairKey = struct
    type ord_key = A.atom * A.atom
    fun compare ((a1,a1'),(a2,a2')) = (case (A.compare (a1,a2))
                                         of LESS => LESS
                                          | GREATER => GREATER
                                          | EQUAL => A.compare (a1',a2'))
  end

  structure AtomKey = struct
    type ord_key = A.atom
    val compare = A.compare
  end

  (* for storing hierarchies *)
  structure H = BinarySetFn (AtomPairKey)
  (* for storing names *)
  structure S = BinarySetFn (AtomKey)
  structure P = SpecParseTree

  fun spname (P.Spec (name,_)) = name

  fun datName (P.Datatype (name,_,_)) = name
  fun specNames (P.Datatype (name,_,P.Specs specs)) = 
    map (fn (P.Spec (name,_)) => name) specs
  fun allNames (g) = datName (g)::specNames (g)

  (* get list of all (specname,constructor,type), including datatype itself *)
  fun allSCT (P.Datatype (name,consts,P.Specs specs)) = let
    fun allConstructors (n,[]) = []
      | allConstructors (n,(P.Const (a,t))::consts) = (n,a,t)::allConstructors (n,consts)
    fun allSpecsConstructors [] = []
      | allSpecsConstructors ((P.Spec (a,consts))::specs) = (allConstructors (a,consts))::allSpecsConstructors (specs)
  in
    List.concat (allConstructors(name,consts)::allSpecsConstructors (specs))
  end

  fun getConsts (consts) = map (fn (P.Const (c,t)) => (c,t)) consts
  fun getDatatypeConsts (P.Datatype (_,consts,_)) = getConsts (consts)
  fun getSpecConsts (n,P.Datatype (_,_,P.Specs specs)) = let
    fun findSpec (n,[]) = Error.fail ["Cannot find spec ",A.toString (n)]
      | findSpec (n,(P.Spec (a,consts))::specs) = 
      if (A.sameAtom (n,a))
        then getConsts (consts)
      else findSpec (n,specs)
  in
    findSpec (n,specs)
  end
    

  (* get list of all (specname,[(constructor,type)]), including datatype itself *)
  fun allCT (P.Datatype (name,consts,P.Specs specs)) = let
    fun allConstructors [] = []
      | allConstructors ((P.Const (a,t))::consts) = (a,t)::allConstructors (consts)
    fun allSpecsConstructors [] = []
      | allSpecsConstructors ((P.Spec (a,consts))::specs) = (a,allConstructors (consts))::allSpecsConstructors (specs)
  in
    (name,allConstructors (consts))::allSpecsConstructors (specs)
  end

  fun ancestors (h,a) = H.foldl (fn ((a1,a2),s) => S.add' (a2,s)) S.empty (H.filter (fn (a1,a2) => A.sameAtom (a,a1)) h)

  fun immAncestors (h,a) = let
    val anc = ancestors (h,a)
  in 
    S.difference (anc,S.foldl (fn (a,s) => S.union (ancestors (h,a),s)) S.empty anc)
  end

  (* find the parent of a node, if any... *)
  fun parent (h,a) = let
    val anc = ancestors (h,a)
  in
    if not (S.numItems (anc) = 1) then 
      Error.fail ["Hierachy is not a tree"]
      else hd (S.listItems (anc))
  end

  fun minimizeHierarchy (g,h) = let
    val names = allNames (g)
    fun addPairs (x,set) = let
      val imm = immAncestors (h,x)
    in 
      S.foldl (fn (y,s) => H.add (s,(x,y))) set imm
    end
  in
    foldl addPairs H.empty names
  end

  fun isSubtype (set,NONE,NONE) = true
    | isSubtype (set,SOME(t1),SOME(t2)) =  let 
        fun isSomeSubtype (set,P.TNamed (n1), P.TNamed (n2)) = (A.sameAtom(n1,n2)) orelse (H.member (set,(n1,n2)))
          | isSomeSubtype (set,P.TProd (t1',t1''),P.TProd (t2',t2'')) = (isSomeSubtype(set,t1',t2') andalso isSomeSubtype(set,t1'',t2''))
          | isSomeSubtype _ = false
      in
        isSomeSubtype (set,t1,t2)
      end
    | isSubtype _ = false

  (* checks if spec1 is a child of spec2 in specialization structure g, 
   based on accumulated information on the hierarchy h *)
  fun isChild (set,spec1,spec2) = let
    val P.Spec (name1,constructors1) = spec1
    val P.Spec (name2,constructors2) = spec2
    fun isSub (P.Const (cname1,typopt1),[]) = false
      | isSub (P.Const (cname1,typopt1), (P.Const (cname2,typopt2))::rest) = 
      if (A.sameAtom(cname1,cname2)) then isSubtype (H.add (set,(spname(spec1),spname(spec2))),typopt1,typopt2)
      else isSub (P.Const (cname1,typopt1),rest)
    fun foreach [] = true
      | foreach (c::cs) = isSub (c,constructors2) andalso foreach(cs)
  in
    foreach (constructors1)
  end


    (* this may break on degenerate hierarchies *)
    (* (1) extract all successors of a node *)
    (* (2) remove from the successors those that are reachable through other successors *)
    (* (3) what's left are the immediate successors *)
    (* the above is kept in a map from atoms (nodes) to sets of successors *)
  fun extractHierarchy (g as P.Datatype (name,constructors,P.Specs specs)) = let
    (* accumulate a set of < relations *)
    fun acc (accset,done,[]) = accset
      | acc (accset,done,(spec::specs)) = let
          (* by default, but should be checked ... *)
          val accset = H.add (accset,(spname(spec),name))
          fun foreach (accset,[]) = accset
             | foreach (accset,donespec::rest) = if (isChild (accset,spec,donespec))
                                            then foreach (H.add (accset,(spname(spec),spname(donespec))),rest)
                                          else if (isChild (accset,donespec,spec))
                                            then foreach (H.add (accset,(spname(donespec),spname(spec))),rest)
                                               else foreach (accset,rest)
          val accset = foreach (accset,done)
        in
             acc (accset,spec::done,specs)
        end
    val h = acc (H.empty,[],specs)
    val h = minimizeHierarchy (g,h)
  in
    print "Hierarchy: \n";
    H.app (fn (a,b) => print (concat ["  ",Atom.toString (a)," < ", Atom.toString (b),"\n"])) h;
    h
  end

  fun sigName (P.Datatype (name,_,_)) = "SPEC_"^(A.toString name)
  fun structName (P.Datatype (name,_,_)) = "Spec_"^(A.toString name)

  fun generateFile (g,h) os = let
    fun out l = TextIO.output (os,(concat l)^"\n")
    fun mkType (a,str) = 
      if (A.sameAtom (a,datName (g))) 
        then str^" t"
      else mkType (parent(h,a),concat ["(",str," enc_",A.toString(a),")"])
    val st = A.toString
    val names = allNames (g)
    val SCT = allSCT(g)
    val CT = allCT (g)
    fun absType (sfx,P.TNamed (a)) = if (List.exists (fn (x) => A.sameAtom (x,a)) names)
                                   then concat ["'a",sfx," A",st(a)]
                                     else st(a)
      | absType (sfx, P.TProd (t1,t2)) = concat [absType ("1",t1)," * ",absType ("2",t2)]
    fun concType (P.TNamed (a)) = if (List.exists (fn (x) => A.sameAtom (x,a)) names)
                                   then concat ["C",st(a)]
                                     else st(a)
      | concType (P.TProd (t1,t2)) = concat [concType (t1)," * ",concType (t2)]
    fun mkConstructorType (n,c,NONE) = out ["  val ",st(c),st(n)," : C",st(n)]
      | mkConstructorType (n,c,SOME (t)) = out ["  val ",st(c),st(n)," : ", absType("",t), " -> C",st(n)]
    fun mkDestructorType (n,cs) = let
      fun mkSingle (c,NONE) = "(unit -> 'b)"
        | mkSingle (c,SOME(t)) = concat ["(",concType (t)," -> 'b)"]
      fun mkTuple [] = Error.fail ["Constructor list for ",st(n)," is empty"]
        | mkTuple [(c,t)] = mkSingle (c,t)
        | mkTuple ((c,t)::cs) = concat [mkSingle (c,t)," * ",mkTuple (cs)]
    in
      out ["  val case",st(n)," : 'a A",st(n)," -> ",mkTuple (cs), " -> 'b"]
    end 
    fun mkDatatype (consts) = let
      fun mkType (P.TNamed (a)) = st(a)
        | mkType (P.TProd (t1,t2)) = concat [mkType (t1), " * ",mkType (t2)]
      fun mkSingle (c,NONE) = st(c)
        | mkSingle (c,SOME(t)) = concat [st(c)," of ",mkType (t)]
      fun mkConsts [] = Error.fail ["Empty constructor list for ",st(datName(g))]
        | mkConsts [c] = mkSingle (c)
        | mkConsts (c::cs) = concat [mkSingle (c)," | ", mkConsts (cs)]
    in
      mkConsts (consts)
    end
    val datConsts = getDatatypeConsts (g)
    fun mkArgs [] = Error.fail ["Empty args list to construct..."]
      | mkArgs [s] = s
      | mkArgs (s::ss) = s^","^mkArgs(ss)
    fun mkSingleCase (c,NONE) = concat [st(c)," => arg",st(c),"()"]
      | mkSingleCase (c,SOME(t)) = concat [st(c),"(x) => arg",st(c)," (x)"]
    fun mkCases []  = Error.fail ["Empty constructor list for ",st(datName(g))]
      | mkCases [c] = mkSingleCase (c)
      | mkCases (c::cs) = concat [mkSingleCase (c)," | ", mkCases (cs)]
    fun mkDestructor (n,cs) = let
      (* FIXME: Relies on the constructors to appear in the same order in 
       * specializations and in the datatype *)
      fun match ([],[]) = []
        | match (d::ds,[]) = "fail"::match (ds,[])
        | match ([],s::ss) = Error.fail ["Constructor list in spec ",st(n)," does not seem to match datatype's"]
        | match ((dc,_)::dcs,s as ((sc,_)::scs)) = if (A.sameAtom (dc,sc))
                                            then ("arg"^(st (dc)))::match(dcs,scs)
                                          else "fail"::match(dcs,s)
    in
      out [" fun case",st(n)," c (",mkArgs (map (fn (c,_) => "arg"^(st c)) cs),") = casef c (",mkArgs (match(datConsts,cs)),")"]
    end
  in
    (* first, the signature *)
    out ["signature ",sigName (g)," = sig"];
    out ["  type 'a t"];
    out [];
    map (fn (s) => out ["  datatype 'a enc_",st (s)," = ENC_",st (s)])  (specNames g);
    out [];
    map (fn (n) => out ["  type 'a A",st (n)," = ",mkType (n,"'a")]) names;
    out [];
    map (fn (n) => out ["  type C",st (n)," = ",mkType (n,"unit")]) names;
    out [];
    map mkConstructorType SCT;
    out [];
    map mkDestructorType CT;
    out [];
    map (fn (n) => out ["  val coerce",st(n)," : 'a A",st(n)," -> C",st(n)]) names;
    out [];
    out ["end"];
    out [];
    (* then, the structure *)
    out ["structure ",structName (g), " :> ",sigName(g)," = struct"];
    out ["  datatype ",st(datName(g))," = ",mkDatatype (datConsts)];
    out [];
    out ["  type 'a t = ",st(datName(g))];
    out [];
    map (fn (s) => out ["  datatype 'a enc_",st (s)," = ENC_",st (s)])  (specNames g);
    out [];
    map (fn (n) => out ["  type 'a A",st (n)," = ",mkType (n,"'a")]) names;
    out [];
    map (fn (n) => out ["  type C",st (n)," = ",mkType (n,"unit")]) names;
    out [];
    map (fn (s,c,_) => out ["  val ",st(c),st(s)," = ",st(c)]) SCT;
    out [];
    out ["  fun fail _ = raise Match"];
    out ["  fun casef c (",mkArgs (map (fn (c,_) => "arg"^(st c)) datConsts),") = case c of"];
    out ["         ",mkCases (datConsts)];
    out [];
    map mkDestructor CT;
    out [];
    out ["  fun coercef v = v"];
    map (fn (n) => out ["  val coerce",st(n)," = coercef"]) names;
    out ["end"]
  end

  fun printInfo (P.Datatype (name,constructors,P.Specs (specs))) = 
    (print ("Datatype = "^(Atom.toString name)^"\n");
     app (fn (P.Spec (name,_)) => print ("Specialization = "^(Atom.toString name)^"\n")) specs)
  
  fun processFile (file) = let 
    val g = Util.withInputFile (file, SpecParser.parse)
    val _ = printInfo (g);
    val hierarchy = extractHierarchy (g)
    val genname = Util.replace_extension (file,"sml")
  in
    Util.withOutputFile (genname,generateFile (g,hierarchy))
  end

end
